//-----------------------------------------------------------------------------
// Copyright (c) 2009-2010 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This application demonstrates how to implement the Phong lighting model for
// a single directional light source using OpenGL 3.1 or a forward compatible
// OpenGL 3.0 rendering context. The shaders can be extended to support
// multiple directional light sources.
//
//-----------------------------------------------------------------------------

#include <windows.h>
#include "application.h"
#include "mathlib.h"

class Demo : public GL3Application
{
public:
	Demo();
    virtual ~Demo();

protected:
	virtual void init(int argc, char *argv[]);
	virtual void render();
	virtual void shutdown();
	virtual void update(float elapsedTimeSec);

private:
	struct UnitCube
	{
		struct Vertex
		{
			float pos[3];
			float normal[3];
			float texCoord[2];
		};

		static const int TOTAL_VERTICES = 36;
		static Vertex vertices[TOTAL_VERTICES];

		const Texture2D *pTexture;

		GLuint shaderProgram;
		GLuint vertexArrayObject;
		GLuint vertexBufferObject;
	};

	void lookAt(const Vector3 &eye, const Vector3 &target, const Vector3 &up);
	const float *normalMatrix();
	void perspective(float fovx, float aspect, float znear, float zfar);
	void rotate(float degrees, float x, float y, float z);
	void unitCubeCreate();
	void unitCubeAddFace(int index, float v1[3], float v2[3], float v3[3], float v4[3],
		                 float n[3], float tc1[2], float tc2[2], float tc3[2], float tc4[2]);
	void unitCubeDestroy();
	void unitCubeRender();

	bool m_paused;
	float m_degrees;
	UnitCube m_cube;
	Matrix4 m_projMatrix;
	Matrix4 m_viewMatrix;
	Matrix4 m_modelMatrix;
	Matrix3 m_normalMatrix;
};

//-----------------------------------------------------------------------------
// Demo class implementation.
//-----------------------------------------------------------------------------

Demo::UnitCube::Vertex Demo::UnitCube::vertices[Demo::UnitCube::TOTAL_VERTICES];

Demo::Demo() : GL3Application("OpenGL 3 Per-Fragment Directional Lighting"), m_paused(false), m_degrees(0.0f)
{
}

Demo::~Demo()
{
}

void Demo::init(int argc, char *argv[])
{
	// Load the shaders.

	if (!(m_cube.shaderProgram = loadShaderProgram("Content\\Shaders\\phong_dir.vert", "Content\\Shaders\\phong_dir.frag")))
	{
		if (programInfoLog().length() > 0)
			throw std::runtime_error(programInfoLog());

		if (shaderInfoLog().length() > 0)
			throw std::runtime_error(shaderInfoLog());
	}

	// Load the textures.

	if (!(m_cube.pTexture = loadTexture2D("Content\\Textures\\classic_floor_tile_diffuse.jpg")))
		throw std::runtime_error("Unable to load texture: classic_floor_tile_diffuse.jpg");

	// Build the cube's geometry.

	unitCubeCreate();

	// Setup the shaders.

	glUseProgram(m_cube.shaderProgram);

	glUniform1i(glGetUniformLocation(m_cube.shaderProgram, "ColorMapSampler"), 0);

	// Setup a directional light that is shining down the negative z world axis.
	glUniform3f(glGetUniformLocation(m_cube.shaderProgram, "DirectionalLight1.direction"), 0.0f, 0.0f, -1.0f);
	glUniform4f(glGetUniformLocation(m_cube.shaderProgram, "DirectionalLight1.ambient"), 1.0f, 1.0f, 1.0f, 1.0f);
	glUniform4f(glGetUniformLocation(m_cube.shaderProgram, "DirectionalLight1.diffuse"), 1.0f, 1.0f, 1.0f, 1.0f);
	glUniform4f(glGetUniformLocation(m_cube.shaderProgram, "DirectionalLight1.specular"), 1.0f, 1.0f, 1.0f, 1.0f);

	// Setup a silver material to show off specular lighting.
	glUniform4f(glGetUniformLocation(m_cube.shaderProgram, "Material1.ambient"), 0.19225f, 0.19225f, 0.19225f, 1.0f);
	glUniform4f(glGetUniformLocation(m_cube.shaderProgram, "Material1.diffuse"), 0.50754f, 0.50754f, 0.50754f, 1.0f);
	glUniform4f(glGetUniformLocation(m_cube.shaderProgram, "Material1.specular"), 0.508273f, 0.508273f, 0.508273f, 1.0f);
	glUniform1f(glGetUniformLocation(m_cube.shaderProgram, "Material1.shininess"), 51.2f);
}

void Demo::render()
{
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	glViewport(0, 0, viewportWidth(), viewportHeight());
	glClearColor(0.3f, 0.5f, 0.9f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	unitCubeRender();
}

void Demo::shutdown()
{
	unitCubeDestroy();
}

void Demo::update(float elapsedTimeSec)
{
	if (keyDown(Keyboard::KEY_LALT) || keyDown(Keyboard::KEY_RALT))
	{
		if (keyPressed(Keyboard::KEY_ENTER))
			toggleFullScreen();
	}

	if (keyPressed(Keyboard::KEY_ESCAPE))
		quit();

	if (keyPressed(Keyboard::KEY_SPACE))
		m_paused = !m_paused;

	if (!m_paused)
	{
		m_degrees += Math::radiansToDegrees(elapsedTimeSec) * 0.5f;
		m_degrees = fmodf(m_degrees, 360.0f);
	}
}

void Demo::lookAt(const Vector3 &eye, const Vector3 &target, const Vector3 &up)
{
	// Builds a look-at style view matrix.
	// This is essentially the same matrix used by gluLookAt().

	Vector3 zAxis = eye - target;
	zAxis.normalize();

	Vector3 xAxis = Vector3::cross(up, zAxis);
	xAxis.normalize();

	Vector3 yAxis = Vector3::cross(zAxis, xAxis);
	yAxis.normalize();

	m_viewMatrix[0][0] = xAxis.x;
	m_viewMatrix[1][0] = xAxis.y;
	m_viewMatrix[2][0] = xAxis.z;
	m_viewMatrix[3][0] = -Vector3::dot(xAxis, eye);

	m_viewMatrix[0][1] = yAxis.x;
	m_viewMatrix[1][1] = yAxis.y;
	m_viewMatrix[2][1] = yAxis.z;
	m_viewMatrix[3][1] = -Vector3::dot(yAxis, eye);

	m_viewMatrix[0][2] = zAxis.x;
	m_viewMatrix[1][2] = zAxis.y;
	m_viewMatrix[2][2] = zAxis.z;
	m_viewMatrix[3][2] = -Vector3::dot(zAxis, eye);

	m_viewMatrix[0][3] = 0.0f;
	m_viewMatrix[1][3] = 0.0f;
	m_viewMatrix[2][3] = 0.0f;
	m_viewMatrix[3][3] = 1.0f;
}

const float *Demo::normalMatrix()
{
	// The normal matrix is the transpose of the inverse of the upper 3x3
	// sub-matrix of the model-view matrix. It is used to transform normals
	// from model space into eye space.

	Matrix4 modelViewMatrix = m_modelMatrix * m_viewMatrix;

	m_normalMatrix[0][0] = modelViewMatrix[0][0];
	m_normalMatrix[0][1] = modelViewMatrix[0][1];
	m_normalMatrix[0][2] = modelViewMatrix[0][2];

	m_normalMatrix[1][0] = modelViewMatrix[1][0];
	m_normalMatrix[1][1] = modelViewMatrix[1][1];
	m_normalMatrix[1][2] = modelViewMatrix[1][2];

	m_normalMatrix[2][0] = modelViewMatrix[2][0];
	m_normalMatrix[2][1] = modelViewMatrix[2][1];
	m_normalMatrix[2][2] = modelViewMatrix[2][2];

	m_normalMatrix = m_normalMatrix.inverse().transpose();

	return &m_normalMatrix[0][0];
}

void Demo::perspective(float fovx, float aspect, float znear, float zfar)
{
	// Construct a projection matrix based on the horizontal field of view
	// 'fovx' rather than the more traditional vertical field of view 'fovy'.

	float e = 1.0f / tanf(Math::degreesToRadians(fovx) / 2.0f);
	float aspectInv = 1.0f / aspect;
	float fovy = 2.0f * atanf(aspectInv / e);
	float xScale = 1.0f / tanf(0.5f * fovy);
	float yScale = xScale / aspectInv;

	m_projMatrix[0][0] = xScale;
	m_projMatrix[0][1] = 0.0f;
	m_projMatrix[0][2] = 0.0f;
	m_projMatrix[0][3] = 0.0f;

	m_projMatrix[1][0] = 0.0f;
	m_projMatrix[1][1] = yScale;
	m_projMatrix[1][2] = 0.0f;
	m_projMatrix[1][3] = 0.0f;

	m_projMatrix[2][0] = 0.0f;
	m_projMatrix[2][1] = 0.0f;
	m_projMatrix[2][2] = (zfar + znear) / (znear - zfar);
	m_projMatrix[2][3] = -1.0f;

	m_projMatrix[3][0] = 0.0f;
	m_projMatrix[3][1] = 0.0f;
	m_projMatrix[3][2] = (2.0f * zfar * znear) / (znear - zfar);
	m_projMatrix[3][3] = 0.0f;
}

void Demo::rotate(float degrees, float x, float y, float z)
{
	m_modelMatrix = Matrix4::createRotate(Vector3(x, y, z), degrees);
}

void Demo::unitCubeCreate()
{
	/*
			   H--------G
			  /|       /|
			 / |      / |
			D--|-----C  |
			|  E-----|--F
			| /      | /
			|/       |/
			A--------B
		Y
		|
		|
		o----X
       /
      /
     Z
			 D-------C
			 |      /|
			 |    /  |
			 |  /    |
			 |/      |
			 A-------B

			      face: ABCD
			 triangles: ABC ACD

		+x face: BFGC triangles: BFG BGC
		-x face: EADH triangles: EAD EDH

		+y face: DCGH triangles: DCG DGH
		-y face: EFBA triangles: EFB EBA

		+z face: ABCD triangles: ABC ACD
		-z face: FEHG triangles: FEH FHG

	*/

	float corners[8][3] =
	{
		-0.5f, -0.5f,  0.5f,	// A
		 0.5f, -0.5f,  0.5f,	// B
		 0.5f,  0.5f,  0.5f,	// C
		-0.5f,  0.5f,  0.5f,	// D

		-0.5f, -0.5f, -0.5f,	// E
		 0.5f, -0.5f, -0.5f,	// F
		 0.5f,  0.5f, -0.5f,	// G
		-0.5f,  0.5f, -0.5f,	// H
	};

	float normals[6][3] =
	{
		 1.0f,  0.0f,  0.0f,	// +x face
		-1.0f,  0.0f,  0.0f,	// -x face
		 0.0f,  1.0f,  0.0f,	// +y face
		 0.0f, -1.0f,  0.0f,	// -y face
		 0.0f,  0.0f,  1.0f,	// +z face
		 0.0f,  0.0f, -1.0f		// -z face
	};

	float texCoords[4][2] =
	{
		0.0f, 0.0f,				// lower left
		1.0f, 0.0f,				// lower right
		1.0f, 1.0f,				// upper right
		0.0f, 1.0f				// upper left
	};

	unitCubeAddFace( 0, corners[1], corners[5], corners[6], corners[2], normals[0], texCoords[0], texCoords[1], texCoords[2], texCoords[3]); // +x face
	unitCubeAddFace( 6, corners[4], corners[0], corners[3], corners[7], normals[1], texCoords[0], texCoords[1], texCoords[2], texCoords[3]); // -x face
	unitCubeAddFace(12, corners[3], corners[2], corners[6], corners[7], normals[2], texCoords[0], texCoords[1], texCoords[2], texCoords[3]); // +y face
	unitCubeAddFace(18, corners[4], corners[5], corners[1], corners[0], normals[3], texCoords[0], texCoords[1], texCoords[2], texCoords[3]); // -y face
	unitCubeAddFace(24, corners[0], corners[1], corners[2], corners[3], normals[4], texCoords[0], texCoords[1], texCoords[2], texCoords[3]); // +z face
	unitCubeAddFace(30, corners[5], corners[4], corners[7], corners[6], normals[5], texCoords[0], texCoords[1], texCoords[2], texCoords[3]); // -z face

	// Create the vertex array object for the cube.
	//
	// The buffer objects that are to be used by the vertex stage of the GL
	// are collected together to form a vertex array object. All state related
	// to the definition of data used by the vertex processor is encapsulated
	// in a vertex array object.
	//
	// Vertex array objects maintain state. Any required vertex and index
	// buffer objects are created and bound to the vertex array object. Then
	// any vertex shader generic shader input variables are mapped to the
	// vertex data stored in the vertex buffer objects bound to the vertex
	// array object. All of this only needs to be done once when using vertex
	// array objects.
	//
	// When we want to draw the cube we simple bind the vertex array object and
	// then call glDrawArrays() to do the drawing. There is no need to define
	// the cube's vertex array format again since that information has already
	// been specified and saved to the vertex array object.
	//
	// See the unitCubeRender() method to see how vertex array objects are used
	// in practice.

	glGenVertexArrays(1, &m_cube.vertexArrayObject);
	glBindVertexArray(m_cube.vertexArrayObject);

	// Create the Vertex Buffer Object for the cube.

	glGenBuffers(1, &m_cube.vertexBufferObject);
	glBindBuffer(GL_ARRAY_BUFFER, m_cube.vertexBufferObject);
	glBufferData(GL_ARRAY_BUFFER, sizeof(UnitCube::vertices), UnitCube::vertices, GL_STATIC_DRAW);

	// GLSL 1.30 (OpenGL 3.0) deprecates varying and attribute shader
	// variables. These are replaced with used-defined generic vertex and
	// fragment shader input and output variables.
	//
	// Map the generic vertex attributes used by the cube's shader program to
	// the cube's vertex data stored in the vertex buffer object.

	glVertexAttribPointer(glGetAttribLocation(m_cube.shaderProgram, "VertexPosition"), 3, GL_FLOAT, GL_FALSE, sizeof(UnitCube::Vertex), BUFFER_OFFSET(0));
	glEnableVertexAttribArray(glGetAttribLocation(m_cube.shaderProgram, "VertexPosition"));

	glVertexAttribPointer(glGetAttribLocation(m_cube.shaderProgram, "VertexNormal"), 3, GL_FLOAT, GL_FALSE, sizeof(UnitCube::Vertex), BUFFER_OFFSET(sizeof(float) * 3));
	glEnableVertexAttribArray(glGetAttribLocation(m_cube.shaderProgram, "VertexNormal"));

	glVertexAttribPointer(glGetAttribLocation(m_cube.shaderProgram, "VertexTexCoord"), 2, GL_FLOAT, GL_FALSE, sizeof(UnitCube::Vertex), BUFFER_OFFSET(sizeof(float) * 6));
	glEnableVertexAttribArray(glGetAttribLocation(m_cube.shaderProgram, "VertexTexCoord"));
}

void Demo::unitCubeAddFace(int index, float v1[3], float v2[3], float v3[3], float v4[3], float n[3], float tc1[2], float tc2[2], float tc3[2], float tc4[2])
{
	UnitCube::Vertex *pVertex = &UnitCube::vertices[index];

	pVertex->pos[0] = v1[0];
	pVertex->pos[1] = v1[1];
	pVertex->pos[2] = v1[2];
	pVertex->normal[0] = n[0];
	pVertex->normal[1] = n[1];
	pVertex->normal[2] = n[2];
	pVertex->texCoord[0] = tc1[0];
	pVertex->texCoord[1] = tc1[1];
	pVertex++;

	pVertex->pos[0] = v2[0];
	pVertex->pos[1] = v2[1];
	pVertex->pos[2] = v2[2];
	pVertex->normal[0] = n[0];
	pVertex->normal[1] = n[1];
	pVertex->normal[2] = n[2];
	pVertex->texCoord[0] = tc2[0];
	pVertex->texCoord[1] = tc2[1];
	pVertex++;

	pVertex->pos[0] = v3[0];
	pVertex->pos[1] = v3[1];
	pVertex->pos[2] = v3[2];
	pVertex->normal[0] = n[0];
	pVertex->normal[1] = n[1];
	pVertex->normal[2] = n[2];
	pVertex->texCoord[0] = tc3[0];
	pVertex->texCoord[1] = tc3[1];
	pVertex++;

	pVertex->pos[0] = v1[0];
	pVertex->pos[1] = v1[1];
	pVertex->pos[2] = v1[2];
	pVertex->normal[0] = n[0];
	pVertex->normal[1] = n[1];
	pVertex->normal[2] = n[2];
	pVertex->texCoord[0] = tc1[0];
	pVertex->texCoord[1] = tc1[1];
	pVertex++;

	pVertex->pos[0] = v3[0];
	pVertex->pos[1] = v3[1];
	pVertex->pos[2] = v3[2];
	pVertex->normal[0] = n[0];
	pVertex->normal[1] = n[1];
	pVertex->normal[2] = n[2];
	pVertex->texCoord[0] = tc3[0];
	pVertex->texCoord[1] = tc3[1];
	pVertex++;

	pVertex->pos[0] = v4[0];
	pVertex->pos[1] = v4[1];
	pVertex->pos[2] = v4[2];
	pVertex->normal[0] = n[0];
	pVertex->normal[1] = n[1];
	pVertex->normal[2] = n[2];
	pVertex->texCoord[0] = tc4[0];
	pVertex->texCoord[1] = tc4[1];
}

void Demo::unitCubeDestroy()
{
	if (m_cube.vertexArrayObject)
	{
		glDeleteVertexArrays(1, &m_cube.vertexArrayObject);
		m_cube.vertexArrayObject = 0;
	}

	if (m_cube.vertexBufferObject)
	{
		glDeleteBuffers(1, &m_cube.vertexBufferObject);
		m_cube.vertexBufferObject = 0;
	}
	
	if (m_cube.shaderProgram)
	{
		glUseProgram(0);
		glDeleteProgram(m_cube.shaderProgram);
		m_cube.shaderProgram = 0;
	}
}

void Demo::unitCubeRender()
{
	perspective(90.0f, float(viewportWidth()) / float(viewportHeight()), 1.0f, 20.0f);
	lookAt(Vector3(0.0f, 0.0f, 4.0f), Vector3(0.0f, 0.0f, 0.0f), Vector3(0.0f, 1.0f, 0.0f));
	rotate(m_degrees, 0.0f, 1.0f, 0.0f);

	glUseProgram(m_cube.shaderProgram);

	// GLSL 1.30 (OpenGL 3.0) deprecates most of the built-in uniform state
	// variables. In GLSL 1.40 (OpenGL 3.1) they have been removed. We now
	// directly supply the projection, view, model, and normal matrices to our
	// shader. OpenGL 3.0 also deprecates fixed-function vertex processing.
	// This means that we also need to calculate the projection, view, model,
	// and normal matrices by ourself.

	glUniformMatrix4fv(glGetUniformLocation(m_cube.shaderProgram, "ProjectionMatrix"), 1, GL_FALSE, &m_projMatrix[0][0]);
	glUniformMatrix4fv(glGetUniformLocation(m_cube.shaderProgram, "ViewMatrix"), 1, GL_FALSE, &m_viewMatrix[0][0]);
	glUniformMatrix4fv(glGetUniformLocation(m_cube.shaderProgram, "ModelMatrix"), 1, GL_FALSE, &m_modelMatrix[0][0]);
	glUniformMatrix3fv(glGetUniformLocation(m_cube.shaderProgram, "NormalMatrix"), 1, GL_FALSE, normalMatrix());
	
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, m_cube.pTexture->texture);

	glBindVertexArray(m_cube.vertexArrayObject);
	glDrawArrays(GL_TRIANGLES, 0, UnitCube::TOTAL_VERTICES);
	glBindVertexArray(0);

	glBindTexture(GL_TEXTURE_2D, 0);
	glUseProgram(0);
}

//-----------------------------------------------------------------------------
// WinMain - application entry point.
//-----------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
    Demo demo;
    
    demo.preferAntiAliasing();
    demo.preferWindowed();

    return demo.run();
}